from unittest import TestCase
from metamake._shell import shell, ShellCommandFailedError


class CoreTestCase(TestCase):
    
    def test_shell_command_raises_exception_when_command_fails(self):
        
        self.assertRaises(ShellCommandFailedError, lambda: shell("bad_command_that_doesnt_exist"))
        


from unittest import TestCase
import sys
from metamake._flags import Flag


def _reset_flag_parser():
    # warning: this uses internal implementation, and will need
    # to be updated accordingly when the implementation changes
    Flag._flag_value_lookup = None
    Flag._flag_explanation_lookup = {}
    Flag._flag_type_lookup = {}
    Flag._flag_allowablevals_lookup = {}

def _add_flag(argstr):
    sys.argv.append(argstr)


class FlagTestCase(TestCase):
    
    def setUp(self):
        _reset_flag_parser()
    
    def tearDown(self):
        _reset_flag_parser()
    
    def test_boolean_flag_detects_true_value(self):
        for value in ["True", "T", "TRUE", "true", "t", "1", "y", "yes", "on"]:
            _reset_flag_parser()
            _add_flag("boolean_flag=%s" % value)
            f = Flag("boolean_flag").explain("blah", convert=bool)
            self.assertEqual(f.value, True)
    
    def test_boolean_flag_detects_false_value(self):
        for value in ["False", "F", "FALSE", "false", "f", "0", "n", "no", "off"]:
            _reset_flag_parser()
            _add_flag("boolean_flag=%s" % value)
            f = Flag("boolean_flag").explain("blah", convert=bool)
            self.assertEqual(f.value, False)
    
    def test_integer_flag_gets_converted_properly(self):
        _add_flag("integer_flag=5")
        f = Flag("integer_flag").explain("blah", convert=int)
        self.assertEqual(f.value, 5)    


from logging import getLogger
# from os import system
import subprocess


class ShellCommandFailedError(Exception):
    """raised whenever a shell command fails"""
    
    stdout = property(lambda self: self.__stdout)
    stderr = property(lambda self: self.__stderr)
    returncode = property(lambda self: self.__returncode)
    
    # TODO: deprecate
    retcode = property(lambda self: self.__returncode)
    
    def __init__(self, cmd, stdout, stderr, returncode):
        Exception.__init__(self, "command '%s' failed with status %s" % (cmd, returncode))
        self.__returncode = returncode
        self.__stdout = stdout
        self.__stderr = stderr


class Result(object):
    
    returncode = property(lambda self: self.__returncode)
    stdout = property(lambda self: self.__stdout)
    stderr = property(lambda self: self.__stderr)
    
    def __init__(self, stdout, stderr, returncode):
        self.__returncode = returncode
        self.__stdout = stdout
        self.__stderr = stderr


def shell(cmd, quiet=False, allow_failure=False):
    """runs a shell command, echoing output and raising
    a SystemExit if the command fails"""
    logger = getLogger("metamake.shell")
    if not quiet:
        logger.info(cmd)
    pipe = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = pipe.communicate()
    # TODO: do line-by-line output
    print stdout.strip()
    print stderr.strip()
    returncode = pipe.returncode
    if returncode is not 0 and not allow_failure:
        raise ShellCommandFailedError(cmd, stdout=stdout, stderr=stderr, returncode=returncode)
    return Result(stdout=stdout, stderr=stderr, returncode=returncode)


import sys
import unittest
import logging

class _UnittestLogHandler(object):
    """acts as a StreamHandler on stdout (does some additional formatting)"""

    def write(self, msg):
        sys.stdout.write("%s" % msg)
    
    def __getattr__(self, name):
        return getattr(sys.stdout, name)

def monitor_log(log_name):
    """monitors a log for unit tests"""
    logger = logging.getLogger( log_name )
    logger.addHandler( logging.StreamHandler( _UnittestLogHandler() ) )
    logger.setLevel( logging.INFO )


###############################################################################
#>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

COVERAGE = [
    '--with-coverage',
    '--cover-package=metamake',
]

FLAGS = ["-v"]

# monitor_log("metamake.filesystem")

#<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
###############################################################################

if __name__ == "__main__":

    import sys
    import os
    import nose

    # prepend pytoss source to the search path to ensure that all pytoss
    # imports happen from the source directory
    sys.path.insert( 0, os.path.join( os.path.dirname(__file__), ".." ) )

    # run nosetests
    nose.core.TestProgram(defaultTest = "tests", argv = (sys.argv + FLAGS))




from StringIO import StringIO
from zlib import compress, decompress
from base64 import b64encode, b64decode
from inspect import getsource
from tempfile import gettempdir
import metamake
from metamake._path import path, path

def serialize_metamake_to_makefile(makefile_filename):
    """given a filename for a makefile, serializes all information
    necessary to that makefile to bootstrap metamake into memory
    and run a Makefile.py"""
    
    # if there is an old makefile, keep the contents in case
    # we need to revert
    if path(makefile_filename).exists():
        old_makefile_file = open(makefile_filename, "r")
        try:
            old_makefile_file_contents = old_makefile_file.read()
            
            # check the version, don't clobber newer metamake versions
            try:
                version_line = old_makefile_file_contents.split("\n")[0]
                if "metamake" in version_line:
                    old_makefile_metamake_version = version_line[11:].strip()
                    if old_makefile_metamake_version > metamake.__version__:
                        # skip this bootstrapping altogether when this version is older
                        return
            except:
                # something was wrong with the version number, assume it is out of date
                pass
            
        finally:
            old_makefile_file.close()
    else:
        old_makefile_file_contents = None
    
    # open the makefile file for writing the serialized code
    new_makefile_file = open(makefile_filename, "w")
    
    # write the version number
    new_makefile_file.write("# metamake %s\n" % metamake.__version__)
    
    try:
        
        # get the metamake module directory
        metamake_dir = path(__file__).dirname()
        
        # iterate over all python files in the module directory
        for python_filename in metamake_dir.listdir("*.py"):
        
            # get the file contents
            raw_contents = open(python_filename).read()
        
            # b64 encode it so that it goes nicely in a one-liner comment
            serialized_contents = b64encode(compress(raw_contents))
        
            # write it out to the makefile
            python_filebasename = python_filename.basename()
            new_makefile_file.write("# %(python_filebasename)s %(serialized_contents)s\n" % locals())
        
    except Exception, e:
        
        new_makefile_file.close()
        
        if old_makefile_file_contents:
            
            # we had an old makefile, revert to it
            reverted_makefile_file = open(makefile_filename)
            try:
                reverted_makefile_file.write(old_makefile_file_contents)
            finally:
                reverted_makefile_file.close()
        else:
            
            # there was no makefile here before, remove the bad one
            path(makefile_filename).remove()
    
    # write the deserialization code into default make targets
    sourcecode = getsource(standalone_deserialize_metamake_from_makefile.func_code)
    oneliner = ";".join([
        "from base64 import b64decode",
        "from imp import load_source",
        "from tempfile import TemporaryFile, gettempdir",
        "tempcodefile = TemporaryFile()",
        "tempcodefile.write(b64decode('%s'))" % b64encode(sourcecode),
        "tempcodefile.seek(0)",
        "makefile_metamake_bootstrapper = load_source('makefile_metamake_bootstrapper', gettempdir(), tempcodefile)",
        "makefile_metamake_bootstrapper.standalone_deserialize_metamake_from_makefile('%s')" % makefile_filename,
    ])
    
    # this is the old way of importing it by manipulating the path
    # load_metamake = "sys.path.insert(0, os.getcwd());import Makefile"
    load_metamake = "import imp;makefile_path=os.path.join(os.path.abspath(os.getcwd()), 'Makefile.py');imp.load_module('Makefile', open(makefile_path, 'r'), makefile_path, ('.py', 'U', 1))"
    new_makefile_file.write('''
BOOTSTRAPMETAMAKE=%(oneliner)s

__default:
	@python -c "$(BOOTSTRAPMETAMAKE);import sys;import os;%(load_metamake)s;import metamake;metamake.run_default_task();"

ls:
	@python -c "$(BOOTSTRAPMETAMAKE);import sys;import os;%(load_metamake)s;import metamake;metamake.list_tasks()"

%%:
	@python -c "$(BOOTSTRAPMETAMAKE);import sys;import os;%(load_metamake)s;import metamake;metamake.run_task('$@', from_makefile=True)"
''' % locals())
    
    # we are done serializing, close the makefile
    new_makefile_file.close()

def standalone_deserialize_metamake_from_makefile(makefile_filename):
    """internal helper definition. deserializes metamake from a makefile
    and places it in the system temp directory (e.g.
    /tmp/metamake_bootstrap/metamake).  It *must* import all of its own
    dependencies since it will be serialized into its own file"""
    
    import sys
    import os
    import os.path
    from zlib import decompress
    from base64 import b64decode
    from tempfile import gettempdir
    
    # set up a path for the metamake
    # TODO: should check the deserialized version of metamake in here, and kill it if its out of date? or are we already accomplishing this elsewhere?
    temp_metamake_import_path = os.path.join(gettempdir(), "metamake_bootstrap")
    temp_metamake_module_path = os.path.join(temp_metamake_import_path, "metamake")
    if not os.path.exists(temp_metamake_import_path):
        os.mkdir(temp_metamake_import_path)
    if not os.path.exists(temp_metamake_module_path):
        os.mkdir(temp_metamake_module_path)
    
    # read the commented lines from the makefile
    makefile_file = open(makefile_filename)
    makefile_lines = []
    try:
        makefile_lines = [line.replace("# ","") for line in makefile_file.readlines() if line.startswith("#") and not line.startswith("# metamake")]
    finally:
        makefile_file.close()
    
    # each line contains the contents of a file, dump them into metamake_bootstrap/metamake
    for line in makefile_lines:
        metamake_subfilename, serialized_source = line.split(" ", 1)
        full_subfile_path_in_temp = os.path.join(temp_metamake_module_path, metamake_subfilename)
        subfile = open(full_subfile_path_in_temp, "w")
        try:
            deserialized_code = decompress(b64decode(serialized_source))
            subfile.write(deserialized_code)
        finally:
            subfile.close()
    
    # add the temp import dir to the system path
    sys.path.insert(0, temp_metamake_import_path)



# TODO: bootstrap needs to check for the metamake-bootstrap header
from __future__ import generators
import sys
from logging import getLogger, StreamHandler, INFO
from metamake._path import LoggablePath as path
from metamake._bootstrap import serialize_metamake_to_makefile
from metamake._flags import Flag
from metamake._dependencies import Dependency
from metamake._shell import shell, ShellCommandFailedError
from metamake._namespaces import namespace, get_current_namespace

__all__ = [
    
    "task",
    "namespace",
    "shell"
    "path",
    "Flag",
    "Dependency",
    "ShellCommandFailedError",
    
    "list_tasks",
    "run_task",
    "run_default_task",
]

__version__ = "1.1.7"

_DEFAULT_STDERR_LOGS = ["metamake.filesystem", "metamake.shell"]
_all_tasks = {}
_ordered_task_keys = []
_first_task_name = None
_default_task_name = None
_did_setup_stderr_logging = False
_bootstrap_filename = None
_defined_flags = False

def bootstrap(bootstrap_filename):
    """call this function to set a bootstrap filename for
    serializing metamake to a Makefile for backwards-compat"""
    global _bootstrap_filename
    _bootstrap_filename = bootstrap_filename

def task(method=None, default=False):
    """this decorator should be used to mark Python functions
    in your Makefile.py as tasks that can be executed on
    the commandline.  You can specify 'default = True' if
    you want that task to be the default."""
    def decorate(method, default):

        global _first_task_name
        global _default_task_name
        
        if _first_task_name == None:
            _first_task_name = method.__name__
        
        if default == True:
            # TODO: how should we handle overwriting of the default task?
            _default_task_name = method.__name__
        
        # index under the function name and namespace
        namespace_list = get_current_namespace()
        if namespace_list:
            lookup_name = ":".join([x.__name__ for x in (namespace_list + [method])])
        else:
            lookup_name = method.__name__
        _all_tasks[lookup_name] = method
        _ordered_task_keys.append(lookup_name)
        
        return method
        
    if method == None:
        def subdecorator(method):
            return decorate(method, default)
        return subdecorator
    else:
        return decorate(method, default)

def list_tasks():
    """lists all tasks that have been defined in your
    Makefile.py file"""
    _define_flags()
    _ensure_stderr_logs_are_active()
    if _bootstrap_filename:
        serialize_metamake_to_makefile(_bootstrap_filename)
    print
    if len(_all_tasks) is 0:
        print "(no tasks defined)"
    else:
        for task_name in _ordered_task_keys:
            task = _all_tasks[task_name]
            shortdoc = task.__doc__
            if not shortdoc:
                shortdoc = "(no documentation)"
            if "\n" in shortdoc:
                shortdoc = shortdoc.split("\n")[0] + "..."
            format_string = "%%(name)-%ss # %%(optional_default_doc)s%%(doc)s" % max([len(name)+1 for name in _all_tasks.keys()])
            print format_string % dict(
                name=task_name,
                doc=shortdoc,
                optional_default_doc={_default_task_name: "(default) "}.get(task_name, ""),
            )
    if len(Flag._flag_explanation_lookup) > 0:
        print
        for flagname, explanation in Flag._flag_explanation_lookup.iteritems():
            print "[%(flagname)s=?] %(explanation)s" % locals()
    print

def run_task(task_name, from_makefile=False):
    """runs a task (by name)"""
    _define_flags()    
    _ensure_stderr_logs_are_active()
    if _bootstrap_filename:
        serialize_metamake_to_makefile(_bootstrap_filename)
    global _all_tasks
    if len(_all_tasks) is 0:
        print >> sys.stderr, "*** no tasks were defined (did you forget to define tasks in the 'Makefile.py' in your working directory?)"
        raise SystemExit(1)
    if task_name in _all_tasks:
        try:
            _all_tasks[task_name]()
        except Exception, e:
            
            # print the error
            print "*** error: %s, %s" % (e.__class__.__name__, str(e))
            
            # show a full traceback if desired
            # if Flag("tracebacks").given and Flag("tracebacks").value is True:
            if "--trace" in sys.argv or (Flag("mmtrace").given and Flag("mmtrace").value == True):
                print "================================"
                import traceback
                traceback.print_tb(sys.exc_info()[2])
            else:
                # tell the user they can get a traceback
                if from_makefile:
                    # use mmtrace=1 as detail flag
                    print "(use mmtrace=1 for more detail)"
                else:
                    # use --trace as detail flag
                    print "(use --trace for more detail)"
                    
            raise SystemExit(1)
    else:
        print >> sys.stderr, "*** there is no task with the name '%s'" % task_name
        raise SystemExit(1)

def run_default_task():
    """runs the task marked as default (using '@task(default=True)')"""
    global _first_task_name
    global _default_task_name
    
    if _default_task_name:
        run_task(_default_task_name)
    
    else:
        # by default, list all tasks if no other default task has been set
        list_tasks()

def _define_flags():
    """helper method to ensure that the global flags are set"""
    global _defined_flags
    if not _defined_flags:
        pass
        # Flag("tracebacks").explain("set to 'true' for full tracebacks on error", convert=bool, allow=[True, False])
        # _defined_flags = True

def _ensure_stderr_logs_are_active():
    """helper method to ensure that file system actions
    are written to stderr"""
    global _did_setup_stderr_logging
    if not _did_setup_stderr_logging:
        stderr_logging_handler = StreamHandler(sys.stderr)
        for logger_name in _DEFAULT_STDERR_LOGS:
            logger = getLogger(logger_name)
            logger.setLevel(INFO)
            logger.addHandler(stderr_logging_handler)
        _did_setup_stderr_logging = True

def metamake_bin():
    """the actual metamake commandline tool (main function)"""
    
    import sys
    
    # allow help from the commandline
    if "--help" in sys.argv:
        print >> sys.stderr, "latest metamake documentation can be found at http://pypi.python.org/pypi/metamake/"
        raise SystemExit(0)

    # the user didn't want help, they are actually trying to exec metamake
    import os
    import imp
    import metamake
    
    try:
        makefile_path = os.path.join(os.path.abspath(os.getcwd()), "Makefile.py")
        imp.load_module("Makefile", open(makefile_path, "r"), makefile_path, ('.py', 'U', 1))

    except (IOError, ImportError), e:
        
        errmsg = str(e)
        if "Makefile" in errmsg or "No such file or directory" in errmsg:
            # this was probably unable to find Makefile.py
            print >> sys.stderr, "*** Makefile.py could not be loaded.  It appears that you have not defined one."
        else:
            # this was probably due to an import from within Makefile.py
            print >> sys.stderr, "*** Makefile.py could not be loaded.  It appears that one of your imports is failing: '%s'." % errmsg
        
        # fail with nonzero error code
        raise SystemExit(1)
    
    # any commandline argument without a '=' in it is a task that we want to run
    tasknames = [arg for arg in sys.argv[1:] if "=" not in arg and arg != "--trace"]
    
    # decide what task(s) to run
    if len(tasknames) is 0:
        
        # no tasknames means we should run the default one
        run_default_task()
        
    elif "ls" in tasknames:
        
        # the 'ls' task is a special task that lists all other tasks
        list_tasks()
    else:
        
        # otherwise, iterate through all task names and run them
        for taskname in tasknames:
            run_task(taskname)



_current_namespace = []

def get_current_namespace():
    return _current_namespace

def namespace(method=None):
    
    def decorate(method):
        global _current_namespace
        old_namespace = list(_current_namespace)
        _current_namespace += [method]

        # define all the subtasks
        method()

        _current_namespace = old_namespace
        return method
        
    if method == None:
        def subdecorator(method):
            return decorate(method)
        return subdecorator
    else:
        return decorate(method)



import os
import sys
from distutils.core import setup

setup(
    name="metamake",
    packages=['metamake'],
    version="1.1.7",
    license="BSD",
    author="Matt Pizzimenti",
    author_email="mjpizz+metamake@gmail.com",
    url="http://pypi.python.org/pypi/metamake/",
    description="Metamake is a dead-simple task-based automation tool written in Python.",
    long_description=open(os.path.join(os.path.dirname(__file__), "README.txt")).read(),
    entry_points='''
[console_scripts]
mm = metamake:metamake_bin
metamake = metamake:metamake_bin
''',
    zip_safe = False,
    )



from unittest import TestCase
import os
from tempfile import gettempdir
import metamake
from metamake._bootstrap import serialize_metamake_to_makefile, standalone_deserialize_metamake_from_makefile
from metamake._path import path

def _remove_bootstrap_dir():
    # warning: this depends on internal implementation
    bootstrap_dir = path(gettempdir()) / "metamake_bootstrap"
    bootstrap_dir.rmtree()


class BootstrapTestCase(TestCase):
    
    def tearDown(self):
        test_makefile = path("Makefile.test")
        if test_makefile.exists():
            test_makefile.remove()
            _remove_bootstrap_dir()
    
    def test_can_serialize_module_to_makefile(self):
        
        serialize_metamake_to_makefile("Makefile")
        standalone_deserialize_metamake_from_makefile("Makefile")
        path("Makefile").remove()
    
    def test_can_use_makefile_bootstrap_to_run_makefilepy(self):
        currentdir = os.path.dirname(__file__)
        serialize_metamake_to_makefile("Makefile")
        self.assert_("# version %s" % metamake.__version__ in open(path("Makefile")).readlines()[0], "first line should be the version")
        self.assert_(os.system("cd %(currentdir)s && make" % locals()) == 0)
        self.assert_(os.system("cd %(currentdir)s && make testA" % locals()) == 0)
        path("Makefile").remove()
    
    def test_makefile_bootstrap_is_generated(self):
        currentdir = os.path.dirname(__file__)
        # serialize to another makefile
        serialize_metamake_to_makefile("Makefile")
        
        # make with that
        self.assert_(os.system("cd %(currentdir)s && make" % locals()) == 0)
        self.assert_(os.system("cd %(currentdir)s && make testA" % locals()) == 0)
        
        # ensure that the test makefile was created as well
        self.assert_(os.path.exists(os.path.join(currentdir, "Makefile.test")))
        self.assert_(os.system("cd %(currentdir)s && make -f Makefile.test" % locals()) == 0)
        self.assert_(os.system("cd %(currentdir)s && make -f Makefile.test testA" % locals()) == 0)
        
        path("Makefile").remove()
